"""Test dataframe utilities."""

import typing
import unittest

import pandas as pd

from curies import Converter
from curies.dataframe import (
    PrefixIndexMethod,
    filter_df_by_curies,
    filter_df_by_prefixes,
    get_df_curies_index,
    get_df_prefixes_index,
    get_filter_df_by_prefixes_index,
)

CONVERTER = Converter.from_prefix_map(
    {
        "a": "https://example.org/a/",
        "b": "https://example.org/b/",
        "c": "https://example.org/c/",
    }
)


class TestDataframe(unittest.TestCase):
    """A test case for dataframe utilities."""

    def test_get_prefix_index(self) -> None:
        """Test getting a prefix index."""
        column = "curie"
        a_curies = [f"a:{i}" for i in range(5)]
        b_curies = [f"b:{i}" for i in range(5)]
        c_curies = [f"c:{i}" for i in range(5)] * 2
        curies = [*a_curies, *b_curies, *c_curies]
        rows = [(curie,) for curie in curies]
        df = pd.DataFrame(rows, columns=[column])

        for method in typing.get_args(PrefixIndexMethod):
            with self.subTest(method=method):
                idx = get_filter_df_by_prefixes_index(
                    df, column, "a", method=method, converter=CONVERTER
                )
                self.assertEqual([0, 1, 2, 3, 4], _rr(idx))

                idx = get_filter_df_by_prefixes_index(
                    df, column, ["a", "b"], method=method, converter=CONVERTER
                )
                self.assertEqual([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], _rr(idx))

                df_a = filter_df_by_prefixes(df, column, "a")
                self.assertEqual(set(a_curies), set(df_a[column]))

                df_ab = filter_df_by_prefixes(df, column, ["a", "b"])
                self.assertEqual({*a_curies, *b_curies}, set(df_ab[column]))

        df_a1 = filter_df_by_curies(df, column, "a:1")
        self.assertEqual({"a:1"}, set(df_a1[column]))

        df_a123 = filter_df_by_curies(df, column, ["a:1", "a:2", "b:1"])
        self.assertEqual({"a:1", "a:2", "b:1"}, set(df_a123[column]))

        dense_prefix_mapping = get_df_prefixes_index(df, column)
        self.assertEqual(
            {
                "a": [0, 1, 2, 3, 4],
                "b": [5, 6, 7, 8, 9],
                "c": [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
            },
            dense_prefix_mapping,
        )

        dense_curie_mapping = get_df_curies_index(df, column)
        self.assertNotIn("a", dense_curie_mapping)
        self.assertNotIn("b", dense_curie_mapping)
        self.assertNotIn("c", dense_curie_mapping)
        self.assertIn("a:0", dense_curie_mapping)
        self.assertEqual([0], dense_curie_mapping["a:0"])
        self.assertEqual([10, 15], dense_curie_mapping["c:0"])


def _rr(series: "pd.Series[bool]") -> list[int]:
    return [index for index, value in enumerate(series) if value]



"""Dataframe operations."""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Collection
from typing import TYPE_CHECKING, Callable, Literal

from typing_extensions import TypeAlias

from .utils import _prefix_from_curie

if TYPE_CHECKING:
    import pandas as pd

    from .api import Converter

__all__ = [
    "PrefixIndexMethod",
    "filter_df_by_curies",
    "filter_df_by_prefixes",
    "get_df_curies_index",
    "get_df_prefixes_index",
    "get_filter_df_by_curies_index",
    "get_filter_df_by_prefixes_index",
]


def _get_prefix_checker(prefix: str | Collection[str]) -> Callable[[str], bool]:
    """Get a function that checks if a CURIE starts with a prefix."""
    if isinstance(prefix, str):
        prefix_with_colon = prefix + ":"

        def _func(curie: str) -> bool:
            return curie.startswith(prefix_with_colon)

    else:
        prefixes_with_colons = {p + ":" for p in prefix}

        def _func(curie: str) -> bool:
            return any(
                curie.startswith(prefix_with_colon) for prefix_with_colon in prefixes_with_colons
            )

    return _func


def _get_prefixes_from_curie_column(
    df: pd.DataFrame, column: int | str, converter: Converter, validate: bool
) -> pd.Series[str]:
    return df[column].map(_get_curie_parser(converter=converter, validate=validate))


def _get_curie_parser(
    *, converter: Converter | None = None, validate: bool = False
) -> Callable[[str], str]:
    # TODO what if it can't parse?
    # TODO handle None?
    # TODO handle invalid CURIEs?

    if not validate:
        return _prefix_from_curie
    elif converter is None:
        raise ValueError("converter is required for validation")
    else:

        def _func(curie: str) -> str:
            reference = converter.parse_curie(curie, strict=True)
            return reference.prefix

    return _func


#: The method for filtering on prefixe
PrefixIndexMethod: TypeAlias = Literal["iterative", "precalculated"]


def get_filter_df_by_prefixes_index(
    df: pd.DataFrame,
    column: str | int,
    prefix: str | Collection[str],
    *,
    method: PrefixIndexMethod | None = None,
    converter: Converter | None = None,
    validate: bool = False,
) -> pd.Series[bool]:
    """Get an index of CURIEs in the given column that start with the prefix(es)."""
    if method == "iterative" or method is None:
        return df[column].map(_get_prefix_checker(prefix))
    elif method == "precalculated":
        if converter is None:  # pragma: no cover
            raise ValueError("a converter is required for method B")
        prefix_series = _get_prefixes_from_curie_column(df, column, converter, validate=validate)
        if isinstance(prefix, str):
            return prefix_series == prefix
        else:
            return prefix_series.isin(prefix)
    else:  # pragma: no cover
        raise ValueError(f"invalid method given: {method}")


def filter_df_by_prefixes(
    df: pd.DataFrame,
    column: str | int,
    prefix: str | Collection[str],
    *,
    method: PrefixIndexMethod | None = None,
    converter: Converter | None = None,
) -> pd.DataFrame:
    """Filter a dataframe based on CURIEs in a given column having a given prefix or set of prefixes.

    :param df: A dataframe
    :param column: The integer index or column name of a column containing CURIEs
    :param prefix: The prefix (given as a string) or collection of prefixes (given as a
        list, set, etc.) to keep
    :param method: The implementation for getting the prefix index
    :param converter: A converter

    :returns: If not in place, return a new dataframe.
    """
    idx = get_filter_df_by_prefixes_index(
        df=df, column=column, prefix=prefix, method=method, converter=converter
    )
    return df[idx]


def get_filter_df_by_curies_index(
    df: pd.DataFrame,
    column: str | int,
    curie: str | Collection[str],
) -> pd.Series[bool]:
    """Get an index of CURIEs in the given column that are the given CURIE(s)."""
    if isinstance(curie, str):
        return df[column] == curie
    else:
        return df[column].isin(set(curie))


def get_df_curies_index(df: pd.DataFrame, column: str | int) -> dict[str, list[int]]:
    """Get a dictionary from CURIEs that appear in the column to the row indexes where they appear."""
    dd: defaultdict[str, list[int]] = defaultdict(list)
    for i, curie in enumerate(df[column]):
        dd[curie].append(i)
    return dict(dd)


def filter_df_by_curies(
    df: pd.DataFrame,
    column: str | int,
    curie: str | Collection[str],
) -> pd.DataFrame:
    """Filter a dataframe based on CURIEs in a given column having a given prefix or set of prefixes.

    :param df: A dataframe
    :param column: The integer index or column name of a column containing CURIEs
    :param curie: The CURIE (given as a string) or collection of CURIEs (given as a
        list, set, etc.) to keep

    :returns: If not in place, return a new dataframe.
    """
    idx = get_filter_df_by_curies_index(df=df, column=column, curie=curie)
    return df[idx]


def get_df_prefixes_index(
    df: pd.DataFrame,
    column: str | int,
    *,
    converter: Converter | None = None,
    validate: bool = False,
) -> dict[str, list[int]]:
    """Get a dictionary from prefixes that appear in the column to the row indexes where they appear."""
    dd: defaultdict[str, list[int]] = defaultdict(list)
    f = _get_curie_parser(converter=converter, validate=validate)
    for i, prefix in enumerate(df[column].map(f)):
        dd[prefix].append(i)
    return dict(dd)



"""Tests for database adapters."""

from __future__ import annotations

import unittest

from curies import Prefix, Reference
from curies.database import get_reference_sa_column, get_reference_sa_composite

prefix = Prefix("hero")
id_1 = "1"
id_2 = "2"
id_3 = "3"
name_1 = "Name 1"
name_2 = "Name 2"
name_3 = "Name 3"


class TestDatabase(unittest.TestCase):
    """Tests for database adapters."""

    def test_sqlalchemy(self) -> None:
        """Test SQLalchemy."""
        from sqlalchemy import Column, Integer, String, create_engine
        from sqlalchemy.orm import DeclarativeBase, Session

        class Base(DeclarativeBase):
            """A base."""

        class MyModel(Base):
            """A SQLAlchemy model that uses a reference."""

            __tablename__ = "my_model"

            id = Column(Integer, primary_key=True)
            reference = get_reference_sa_column()
            name = Column(String)

        engine = create_engine("sqlite://")
        Base.metadata.create_all(engine)

        model_1 = MyModel(reference=Reference(prefix=prefix, identifier=id_1), name=name_1)
        model_2 = MyModel(reference=Reference(prefix=prefix, identifier=id_2), name=name_2)
        model_3 = MyModel(reference=Reference(prefix=prefix, identifier=id_3), name=name_3)

        with Session(engine) as session:
            session.add(model_1)
            session.add(model_2)
            session.add(model_3)
            session.commit()

        # Test querying with reconstitution
        with Session(engine) as session:
            result = (
                session.query(MyModel)
                .filter(MyModel.reference == Reference(prefix=prefix, identifier=id_1))
                .one()
            )

        self.assertIsInstance(result.reference, Reference)
        self.assertEqual(prefix, result.reference.prefix)
        self.assertEqual(id_1, result.reference.identifier)
        self.assertEqual(name_1, result.name)

    def test_sqlalchemy_composite(self) -> None:
        """Test SQLAlchemy composite models."""
        from sqlalchemy import Column, Integer, String, create_engine
        from sqlalchemy.orm import DeclarativeBase, Session

        class Base(DeclarativeBase):
            """A base."""

        class MyModel(Base):
            """A model for testing composite references."""

            __tablename__ = "my_model"

            id = Column(Integer, primary_key=True)

            reference_prefix = Column(String, nullable=False)
            reference_identifier = Column(String, nullable=False)
            name = Column(String, nullable=False)

            reference = get_reference_sa_composite(reference_prefix, reference_identifier)

        engine = create_engine("sqlite://")
        Base.metadata.create_all(engine)

        model_1 = MyModel(reference=Reference(prefix=prefix, identifier=id_1), name=name_1)
        model_2 = MyModel(reference=Reference(prefix=prefix, identifier=id_2), name=name_2)
        model_3 = MyModel(reference=Reference(prefix=prefix, identifier=id_3), name=name_3)

        with Session(engine) as session:
            session.add(model_1)
            session.add(model_2)
            session.add(model_3)
            session.commit()

        # Test querying with reconstitution
        with Session(engine) as session:
            result = (
                session.query(MyModel)
                .filter(MyModel.reference == Reference(prefix=prefix, identifier=id_1))
                .one()
            )

        self.assertIsInstance(result.reference, Reference)
        self.assertEqual(prefix, result.reference.prefix)
        self.assertEqual(id_1, result.reference.identifier)
        self.assertEqual(name_1, result.name)

    def test_sqlmodel(self) -> None:
        """Test SQLModel."""
        from sqlmodel import Field, Session, SQLModel, create_engine, select

        class Model(SQLModel, table=True):
            """A class with a reference."""

            id: int | None = Field(default=None, primary_key=True)
            reference: Reference = Field(sa_column=get_reference_sa_column())
            optional_reference: Reference | None = Field(sa_column=get_reference_sa_column())
            name: str

        model_1 = Model(reference=Reference(prefix=prefix, identifier=id_1), name=name_1)
        model_2 = Model(reference=Reference(prefix=prefix, identifier=id_2), name=name_2)
        model_3 = Model(
            reference=Reference(prefix=prefix, identifier=id_3),
            optional_reference=Reference(prefix=prefix, identifier=id_3),
            name=name_3,
        )

        engine = create_engine("sqlite://")

        SQLModel.metadata.create_all(engine)

        with Session(engine) as session:
            session.add(model_1)
            session.add(model_2)
            session.add(model_3)
            session.commit()

        # Test querying with reconstitution
        with Session(engine) as session:
            statement = select(Model).where(Model.name == name_2)
            result_1 = session.exec(statement).first()

        if result_1 is None:
            raise TypeError  # this shouldn't be possible
        self.assertIsInstance(result_1.reference, Reference)
        self.assertEqual(prefix, result_1.reference.prefix)
        self.assertEqual(id_2, result_1.reference.identifier)
        self.assertEqual(name_2, result_1.name)

        # Test querying using a string on the reference
        with Session(engine) as session:
            statement = select(Model).where(Model.reference == f"{prefix}:{id_3}")
            result_4 = session.exec(statement).first()

        if result_4 is None:
            raise TypeError  # this shouldn't be possible
        self.assertIsInstance(result_4.reference, Reference)
        self.assertEqual(prefix, result_4.reference.prefix)
        self.assertEqual(id_3, result_4.reference.identifier)
        self.assertEqual(name_3, result_4.name)
        self.assertIsNotNone(result_4.optional_reference)

        # Tests looking up a reference that's missing
        with Session(engine) as session:
            statement = select(Model).where(
                Model.reference == Reference(prefix="nope", identifier="nope")
            )
            self.assertIsNone(session.exec(statement).one_or_none())

        # Tests looking a reference that's there
        with Session(engine) as session:
            statement = select(Model).where(
                Model.reference == Reference(prefix=prefix, identifier=id_1)
            )
            result_3 = session.exec(statement).first()

        if result_3 is None:
            raise TypeError  # this shouldn't be possible
        self.assertIsInstance(result_3.reference, Reference)
        self.assertEqual(prefix, result_3.reference.prefix)
        self.assertEqual(id_1, result_3.reference.identifier)
        self.assertEqual(name_1, result_3.name)



"""Database adapters for :mod:`curies`.

Using :mod:`sqlmodel`
=====================

SQLModel is a joint abstraction over :mod:`pydantic` and :mod:`sqlalchemy`. If you want
to use :class:`curies.Reference` within a SQLModel class, you can do so by setting the
SQLAlchemy column type through the ``sa_column`` keyword argument to the results
returned by :func:`get_reference_sa_column`,

For example, this can be used to model a semantic triple, which has a subject reference,
predicate reference, and object reference like in the following:

.. code-block:: python

    from curies import Reference
    from curies.database import get_reference_sa_column
    from sqlmodel import Field, Session, SQLModel, create_engine, select


    class Edge(SQLModel, table=True):
        id: int | None = Field(default=None, primary_key=True)
        subject: Reference = Field(sa_column=get_reference_sa_column())
        predicate: Reference = Field(sa_column=get_reference_sa_column())
        object: Reference = Field(sa_column=get_reference_sa_column())


    e1 = Edge(subject="CHEBI:135122", predicate="skos:exactMatch", object="mesh:C073738")
    e2 = Edge(subject="CHEBI:135125", predicate="skos:exactMatch", object="mesh:C073260")

    engine = create_engine("sqlite://")
    SQLModel.metadata.create_all(engine)

    # Add edges to the database
    with Session(engine) as session:
        session.add_all([e1, e2])
        session.commit()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(Edge.subject == "CHEBI:135122")
        edges = session.exec(statement).all()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(
            Edge.subject == Reference(prefix="CHEBI", identifier="135125")
        )
        edges = session.exec(statement).all()

If you want to have a field that has an optional reference, you can do it as expected
using the same column type, such as for the ``author`` field below:

.. code-block:: python

    class Edge(SQLModel, table=True):
        id: int | None = Field(default=None, primary_key=True)
        subject: Reference = Field(sa_column=get_reference_sa_column())
        predicate: Reference = Field(sa_column=get_reference_sa_column())
        object: Reference = Field(sa_column=get_reference_sa_column())
        author: Reference | None = Field(sa_column=get_reference_sa_column())

Using :mod:`sqlalchemy`
=======================

SQLAlchemy is a combine high- and mid-level database abstraction layer and
object-relational mapping. It has more opportunities for configuration over SQLModel.

The following example re-formulates the above example for SQLModel using the slightly
more direct SQLAlchemy formulation:

.. code-block:: python

    from curies import Reference
    from sqlalchemy import Column, Integer, String, create_engine
    from sqlalchemy.orm import DeclarativeBase, Session


    class Base(DeclarativeBase):
        pass


    class Edge(Base):
        __tablename__ = "edge"

        id = Column(Integer, primary_key=True)

        subject = get_reference_sa_column()
        predicate = get_reference_sa_column()
        object = get_reference_sa_column()


    e1 = Edge(subject="CHEBI:135122", predicate="skos:exactMatch", object="mesh:C073738")
    e2 = Edge(subject="CHEBI:135125", predicate="skos:exactMatch", object="mesh:C073260")

    engine = create_engine("sqlite://")
    Base.metadata.create_all(engine)

    # Add edges to the database
    with Session(engine) as session:
        session.add_all([e1, e2])
        session.commit()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(Edge.subject == "CHEBI:135122")
        edges = session.exec(statement).all()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(
            Edge.subject == Reference(prefix="CHEBI", identifier="135125")
        )
        edges = session.exec(statement).all()

In the following example, SQLAlchemy is used with a composite column, where the prefix
and identifier are given their own column, then a :class:`sqlalchemy.Composite` object
is used to tell the ORM to map both the prefix and identifier together into a single
column, that exposes an appropriate :class:`curies.Reference` class.

.. code-block:: python

    from curies import Reference
    from sqlalchemy import Column, Integer, String, create_engine
    from sqlalchemy.orm import DeclarativeBase, Session


    class Base(DeclarativeBase):
        pass


    class Edge(Base):
        __tablename__ = "edge"

        id = Column(Integer, primary_key=True)

        subject_prefix = Column(String, nullable=False)
        subject_identifier = Column(String, nullable=False)
        predicate_prefix = Column(String, nullable=False)
        predicate_identifier = Column(String, nullable=False)
        object_prefix = Column(String, nullable=False)
        object_identifier = Column(String, nullable=False)

        subject = get_reference_sa_composite(subject_prefix, subject_identifier)
        predicate = get_reference_sa_composite(predicate_prefix, predicate_identifier)
        object = get_reference_sa_composite(object_prefix, object_identifier)


    e1 = Edge(subject="CHEBI:135122", predicate="skos:exactMatch", object="mesh:C073738")
    e2 = Edge(subject="CHEBI:135125", predicate="skos:exactMatch", object="mesh:C073260")

    engine = create_engine("sqlite://")
    Base.metadata.create_all(engine)

    # Add edges to the database
    with Session(engine) as session:
        session.add_all([e1, e2])
        session.commit()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(Edge.subject == "CHEBI:135122")
        edges = session.exec(statement).all()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(
            Edge.subject == Reference(prefix="CHEBI", identifier="135125")
        )
        edges = session.exec(statement).all()
"""

from __future__ import annotations

from typing import Any, ClassVar

import sqlalchemy
from sqlalchemy import Column
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.orm import Composite, composite
from sqlalchemy.types import TEXT, TypeDecorator

from curies import Reference

__all__ = [
    "SAReferenceTypeDecorator",
    "get_reference_sa_column",
    "get_reference_sa_composite",
]


class SAReferenceTypeDecorator(TypeDecorator[Reference]):
    """A SQLAlchemy type decorator for a :mod:`curies.Reference`."""

    impl = TEXT
    #: Set SQLAlchemy caching to true
    cache_ok: ClassVar[bool] = True  # type:ignore

    def process_bind_param(self, value: str | Reference | None, dialect: Dialect) -> str | None:
        """Convert the Python object into a database value."""
        if value is None:
            return None
        if isinstance(value, str):
            return value
        return value.curie

    def process_result_value(self, value: str | None, dialect: Dialect) -> Reference | None:
        """Convert the database value into a Python object."""
        if value is None:
            return None
        return Reference.from_curie(value)

    # TODO what about process literal param?


def get_reference_sa_column(*args: Any, **kwargs: Any) -> sqlalchemy.Column[Reference]:
    """Get a SQLAlchemy column with the type decorator for a :mod:`curies.Reference`.

    :param args: positional arguments, passed to :class:`sqlalchemy.Column`
    :param kwargs: keyword arguments, passed to :class:`sqlalchemy.Column`

    :returns: A column object, parametrized with :class:`curies.Reference`

    For example, this can be used to model a semantic triple, which has a subject
    reference, predicate reference, and object reference like in the following:

    .. code-block:: python

        from curies import Reference
        from curies.database import get_reference_sa_column
        from sqlmodel import Field, SQLModel


        class Edge(SQLModel, table=True):
            id: int | None = Field(default=None, primary_key=True)
            subject: Reference = Field(sa_column=get_reference_sa_column())
            predicate: Reference = Field(sa_column=get_reference_sa_column())
            object: Reference = Field(sa_column=get_reference_sa_column())
    """
    return sqlalchemy.Column(SAReferenceTypeDecorator(), *args, **kwargs)


class _ReferenceAdapter(Reference):
    """A wrapper for SQLAlchemy for usage in composite()."""

    def __init__(self, prefix: str, identifier: str) -> None:
        """Initialize the SQLAlchemy model."""
        super().__init__(prefix=prefix, identifier=identifier)


def get_reference_sa_composite(
    prefix_column: Column[str], identifier_column: Column[str], **kwargs: Any
) -> Composite[Reference]:
    """Get a composite for a reference.

    :param prefix_column: The column for the reference's prefix
    :param identifier_column: The column for the reference's identifier
    :param kwargs: keyword arguments passed to :func:`sqlalchemy.orm.composite`

    :returns: A Composite object for a reference

    .. code-block:: python

        from curies import Reference
        from curies.database import get_reference_sa_composite
        from sqlalchemy import Column
        from sqlalchemy.orm import DeclarativeBase


        class Base(DeclarativeBase):
            pass


        class Edge(Base):
            __tablename__ = "edge"

            id = Column(Integer, primary_key=True)

            subject_prefix = Column(String, nullable=False)
            subject_identifier = Column(String, nullable=False)
            predicate_prefix = Column(String, nullable=False)
            predicate_identifier = Column(String, nullable=False)
            object_prefix = Column(String, nullable=False)
            object_identifier = Column(String, nullable=False)

            subject = get_reference_sa_composite(subject_prefix, subject_identifier)
            predicate = get_reference_sa_composite(predicate_prefix, predicate_identifier)
            object = get_reference_sa_composite(object_prefix, object_identifier)
    """
    return composite(_ReferenceAdapter, prefix_column, identifier_column, **kwargs)
